/*
    Project: TCGINA
    Version: V1.23
    File   : setupdlg.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <shellapi.h>
#include <dbt.h>
#include <sddl.h>
#include <tchar.h>
#include <tc.h>
#include "setupdlg.h"
#include "misc.h"
#include "project.h"

#ifndef TC_DEBUG
#  define DISPLAY_ONLY_KNOWN_TCGINA_FILES
#endif

static BOOL CopyFileSecurityInformation(LPCTSTR pszSrcFileName,
  LPCTSTR pszDestFileName, SECURITY_INFORMATION si)
{
    PSECURITY_DESCRIPTOR pSD;
    DWORD                cbSize;
    DWORD                cbNewSize;
    BOOL                 fResult;

    cbSize = 0;
    GetFileSecurity(pszSrcFileName, si, NULL, 0, &cbSize);
    pSD = malloc(cbSize + 1);
    if (pSD == NULL) {
        return FALSE;
    }

    fResult = FALSE;
    if (GetFileSecurity(pszSrcFileName, si, pSD, cbSize, &cbNewSize) &&
      cbNewSize <= cbSize) {
        if (SetFileSecurity(pszDestFileName, si, pSD)) {
            fResult = TRUE;
        }
    }

    free(pSD);

    return fResult;
}

static BOOL CopyFileSecurity(LPCTSTR pszSrcFileName, LPCTSTR pszDestFileName)
{
    BOOL fResult;

    fResult = TRUE;
    if (!CopyFileSecurityInformation(pszSrcFileName, pszDestFileName,
      OWNER_SECURITY_INFORMATION)) {
        fResult = FALSE;
    }
    if (!CopyFileSecurityInformation(pszSrcFileName, pszDestFileName,
      GROUP_SECURITY_INFORMATION)) {
        fResult = FALSE;
    }
    if (!CopyFileSecurityInformation(pszSrcFileName, pszDestFileName,
      DACL_SECURITY_INFORMATION)) {
        fResult = FALSE;
    }
    if (!CopyFileSecurityInformation(pszSrcFileName, pszDestFileName,
      SACL_SECURITY_INFORMATION)) {
#if 0
        fResult = FALSE;
#endif
    }
    if (!CopyFileSecurityInformation(pszSrcFileName, pszDestFileName,
      GROUP_SECURITY_INFORMATION)) {
        fResult = FALSE;
    }
    if (!CopyFileSecurityInformation(pszSrcFileName, pszDestFileName,
      PROTECTED_DACL_SECURITY_INFORMATION)) {
        fResult = FALSE;
    }
    if (!CopyFileSecurityInformation(pszSrcFileName, pszDestFileName,
      PROTECTED_SACL_SECURITY_INFORMATION)) {
        fResult = FALSE;
    }
    if (!CopyFileSecurityInformation(pszSrcFileName, pszDestFileName,
      UNPROTECTED_DACL_SECURITY_INFORMATION)) {
        fResult = FALSE;
    }
    if (!CopyFileSecurityInformation(pszSrcFileName, pszDestFileName,
      UNPROTECTED_SACL_SECURITY_INFORMATION)) {
        fResult = FALSE;
    }

    return fResult;
}

static BOOL CopyPathSecurity(LPCTSTR pszSrcFileName,
  LPCTSTR pszDestFileName)
{
    TCHAR           szSrcFileName[MAX_PATH];
    TCHAR           szDestFileName[MAX_PATH];
    LPTSTR          pszSrcFilePart;
    LPTSTR          pszDestFilePart;
    DWORD           cchSrcPathSize;
    DWORD           cchDestPathSize;
    DWORD           cchMaxFileSize;
    HANDLE          hFindFile;
    WIN32_FIND_DATA fd;
    BOOL            fResult;

    fResult = CopyFileSecurity(pszSrcFileName, pszDestFileName);

    cchSrcPathSize = (DWORD)_tcslen(pszSrcFileName);
    cchDestPathSize = (DWORD)_tcslen(pszDestFileName);
    if (cchSrcPathSize >= MAX_PATH - 4 || cchDestPathSize >= MAX_PATH - 1) {
        return FALSE;
    }
    cchMaxFileSize = (cchSrcPathSize >= cchDestPathSize) ?
      (MAX_PATH - cchSrcPathSize - 2) : (MAX_PATH - cchDestPathSize - 2);

    memcpy(szSrcFileName, pszSrcFileName, cchSrcPathSize*sizeof(TCHAR));
    _tcscpy(szSrcFileName + cchSrcPathSize, TEXT("\\*.*"));
    memcpy(szDestFileName, pszDestFileName, cchDestPathSize*sizeof(TCHAR));
    _tcscpy(szDestFileName + cchSrcPathSize, TEXT("\\"));

    pszSrcFilePart = szSrcFileName + cchSrcPathSize + 1;
    pszDestFilePart = szDestFileName + cchDestPathSize + 1;

    hFindFile = FindFirstFile(szSrcFileName, &fd);
    if (hFindFile != INVALID_HANDLE_VALUE) {
        do {
            if (fd.cFileName[0] == TEXT('.') &&
              (fd.cFileName[1] == TEXT('\0') ||
              (fd.cFileName[1] == TEXT('.') &&
              (fd.cFileName[2] == TEXT('\0'))))) {
                continue;
            }
            if (_tcslen(fd.cFileName) > cchMaxFileSize) {
                fResult = FALSE;
                continue;
            }
            _tcscpy(pszSrcFilePart, fd.cFileName);
            _tcscpy(pszDestFilePart, fd.cFileName);
            if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
                if (!CopyPathSecurity(szSrcFileName,
                  szDestFileName)) {
                    fResult = FALSE;
                }
            } else {
                if (!CopyFileSecurity(szSrcFileName,
                  szDestFileName)) {
                    fResult = FALSE;
                }
            }
        } while (FindNextFile(hFindFile, &fd));
        FindClose(hFindFile);
    }

    return fResult;
}

static BOOL ProtectFile(LPCTSTR pszFileName)
{
    LPVOID lpSecurityDescriptor;
    BOOL fResult = FALSE;

    if (ConvertStringSecurityDescriptorToSecurityDescriptor(
      TEXT("D:")
      TEXT("(A;OICI;GA;;;SY)")
      TEXT("(A;OICI;GA;;;BA)")
      TEXT("(A;OICI;GRGX;;;PU)")
      TEXT("(A;OICI;GRGX;;;BU)")
      TEXT(""),
      SDDL_REVISION_1, &lpSecurityDescriptor, NULL)) {
        if (SetFileSecurity(pszFileName,
          DACL_SECURITY_INFORMATION, lpSecurityDescriptor)) {
            fResult = TRUE;
        }
        LocalFree(lpSecurityDescriptor);
    }

    return fResult;
}

static BOOL Install(PSETUP_DIALOG_DATA psdd)
{
    FILE_INFO  fi;
    DWORD      dwIndex;
    PFILE_INFO pfiCurrGinaDLL;
    PFILE_INFO pfiNewGinaDLL;
    TCHAR      szGinaFileName[MAX_PATH];
    TCHAR      szMsGinaFileName[MAX_PATH];
    LPTSTR     pszFilePart;
    unsigned   i;

    dwIndex = (DWORD)SendMessage(psdd->hwndInstallComboBox,
      CB_GETCURSEL, 0, 0);
    if (dwIndex >= psdd->dwNewGinaDLLCount) {
        return FALSE;
    }
    pfiNewGinaDLL = psdd->pfiNewGinaDLL + dwIndex;
    pfiCurrGinaDLL = &psdd->fiGinaDLL;

    GetSystemDirectory(szGinaFileName, MAX_PATH);
    i = (unsigned) _tcslen(szGinaFileName);
    if (i > MAX_PATH - 20) {
        return FALSE;
    }
    szGinaFileName[i] = TEXT('\\');
    pszFilePart = szGinaFileName + i + 1;

    GetSystemDirectory(szMsGinaFileName, MAX_PATH);
    i = (unsigned) _tcslen(szMsGinaFileName);
    if (i > MAX_PATH - 20) {
        return FALSE;
    }
    _tcscpy(szMsGinaFileName + i, TEXT("\\MSGINA.DLL"));

    /* Install TCGINA.DLL */
    for (i = 1; i < 100; i++) {
        if (i == 1) {
            wsprintf(pszFilePart, TEXT("tcgina.dll"));
        } else {
            wsprintf(pszFilePart, TEXT("tcgina%u.dll"), i);
        }
        GetFileInfo(&fi, szGinaFileName);
        if (fi.fExists && !IsTcginaDLL(&fi)) {
            continue;
        }
        if (fi.fInUse) {
            continue;
        }
        if (CopyFile(pfiNewGinaDLL->fd.cFileName, szGinaFileName, FALSE)) {
            break;
        }
        DeleteFile(szGinaFileName);
    }
    if (i == 100) {
        return FALSE;
    }

    if (!CopyFileSecurityInformation(szMsGinaFileName, szGinaFileName,
      DACL_SECURITY_INFORMATION)) {
        ProtectFile(szGinaFileName);
    }

    if (!IsTcginaDLL(pfiCurrGinaDLL)) {
        /* Clone GinaDLL registry value */
        if (!RegCloneValue(HKEY_LOCAL_MACHINE, WINLOGON_KEY, TEXT("GinaDLL"),
          HKEY_LOCAL_MACHINE, TCGINA_KEY, TEXT("GinaDLL"))) {
            DeleteFile(szGinaFileName);
            return FALSE;
        }
    } else {
        /* Check whether a cascaded GinaDLL exists */
        TCHAR szCascadedGinaDLL[MAX_PATH];
        TCHAR szFullPathName[MAX_PATH];
        *szCascadedGinaDLL = TEXT('\0');
        TC_GetRegistryString(HKEY_LOCAL_MACHINE, TCGINA_KEY, TEXT("GinaDLL"),
          szCascadedGinaDLL, MAX_PATH);
        if (*szCascadedGinaDLL) {
            FILE_INFO fiCascadedGinaDLL;
            SearchPath(NULL, szCascadedGinaDLL, TEXT(".DLL"), MAX_PATH,
              szFullPathName, &pszFilePart);
            GetFileInfo(&fiCascadedGinaDLL, szFullPathName);
            if (fiCascadedGinaDLL.fExists || IsTcginaDLL(&fiCascadedGinaDLL)) {
                if (!RegRemoveValue(HKEY_LOCAL_MACHINE, TCGINA_KEY,
                  TEXT("GinaDLL"))) {
                    DeleteFile(szGinaFileName);
                    return FALSE;
                }
            }
        }
    }
    if (!RegSetString(HKEY_LOCAL_MACHINE, WINLOGON_KEY, TEXT("GinaDLL"),
      REG_SZ, szGinaFileName)) {
        DeleteFile(szGinaFileName);
        return FALSE;
    }

    return TRUE;
}

static BOOL UninstallUser(PUSER_INFO_ITEM puii)
{
    TCHAR szProfileKey[MAX_PATH];
    TCHAR szTcginaKey[2*MAX_PATH];

    TC_CopyString(szProfileKey, countof(szProfileKey), PROFILELIST_KEY);
    TC_AppendString(szProfileKey, countof(szProfileKey), TEXT("\\"));
    TC_AppendString(szProfileKey, countof(szProfileKey), puii->szSID);

    TC_CopyString(szTcginaKey, countof(szTcginaKey), TCGINA_KEY);
    TC_AppendString(szTcginaKey, countof(szTcginaKey), TEXT("\\"));
    TC_AppendString(szTcginaKey, countof(szTcginaKey), puii->szUserName);

    if (!RegCopyValue(HKEY_LOCAL_MACHINE, szTcginaKey,
      TEXT("ProfileImagePath"), HKEY_LOCAL_MACHINE, szProfileKey,
      TEXT("ProfileImagePath"))) {
        return FALSE;
    }
    if (RegValueExists(HKEY_LOCAL_MACHINE, szProfileKey,
      TEXT("ProfileLoadTimeHigh"))) {
        RegSetInteger(HKEY_LOCAL_MACHINE, szProfileKey,
          TEXT("ProfileLoadTimeHigh"), 0);
    }
    if (RegValueExists(HKEY_LOCAL_MACHINE, szProfileKey,
      TEXT("ProfileLoadTimeLow"))) {
        RegSetInteger(HKEY_LOCAL_MACHINE, szProfileKey,
          TEXT("ProfileLoadTimeLow"), 0);
    }

    RegRemoveValue(HKEY_LOCAL_MACHINE, szTcginaKey,
      TEXT("ProfileImagePath"));
    RegRemoveValue(HKEY_LOCAL_MACHINE, szTcginaKey,
      TEXT("EncryptedProfileImagePath"));
    RegRemoveValue(HKEY_LOCAL_MACHINE, szTcginaKey,
      TEXT("SID"));
    RegRemoveValue(HKEY_LOCAL_MACHINE, szTcginaKey,
      TEXT("TCDrive"));
    RegRemoveValue(HKEY_LOCAL_MACHINE, szTcginaKey,
      TEXT("TCFileName"));
    if (TC_GetRegistryInteger(HKEY_LOCAL_MACHINE, szTcginaKey,
      TEXT("DisableRedirectionDetection"), 0) == 2) {
        RegRemoveValue(HKEY_LOCAL_MACHINE, szTcginaKey,
          TEXT("DisableRedirectionDetection"));
    }
    if (RegKeyIsEmpty(HKEY_LOCAL_MACHINE, szTcginaKey)) {
        RegRemoveKey(HKEY_LOCAL_MACHINE, szTcginaKey);
        if (RegKeyIsEmpty(HKEY_LOCAL_MACHINE, TCGINA_KEY)) {
            RegRemoveKey(HKEY_LOCAL_MACHINE, TCGINA_KEY);
            if (RegKeyIsEmpty(HKEY_LOCAL_MACHINE, TCGINA_ROOT_KEY)) {
                RegRemoveKey(HKEY_LOCAL_MACHINE, TCGINA_ROOT_KEY);
            }
        }
    }

    return TRUE;
}

static BOOL Uninstall(PSETUP_DIALOG_DATA psdd, PBOOL pfRebootRequired)
{
    PFILE_INFO_LIST pfil;

    if (IsTcginaDLL(&psdd->fiGinaDLL)) {
        BOOL   fCascadedGinaDLLExists;
        TCHAR  szCascadedGinaDLL[MAX_PATH];
        TCHAR  szFullPathName[MAX_PATH];
        LPTSTR pszFilePart;

        /* Check whether a cascaded GinaDLL exists */
        fCascadedGinaDLLExists = FALSE;
        *szCascadedGinaDLL = TEXT('\0');
        TC_GetRegistryString(HKEY_LOCAL_MACHINE, TCGINA_KEY, TEXT("GinaDLL"),
            szCascadedGinaDLL, MAX_PATH);
        if (*szCascadedGinaDLL) {
            SearchPath(NULL, szCascadedGinaDLL, TEXT(".DLL"), MAX_PATH,
            szFullPathName, &pszFilePart);
            if (FileExists(szFullPathName)) {
                fCascadedGinaDLLExists = TRUE;
            }
        }
        /* Unregister TCGINA */
        if (fCascadedGinaDLLExists) {
            if (!RegCloneValue(HKEY_LOCAL_MACHINE, TCGINA_KEY, TEXT("GinaDLL"),
            HKEY_LOCAL_MACHINE, WINLOGON_KEY, TEXT("GinaDLL"))) {
                return FALSE;
            }
        } else {
            if (!RegRemoveValue(HKEY_LOCAL_MACHINE, WINLOGON_KEY,
              TEXT("GinaDLL"))) {
                return FALSE;
            }
        }
        UninstallFile(psdd->fiGinaDLL.fd.cFileName, pfRebootRequired);
    }

    /* Remove files */
    for (pfil = psdd->pfilFirstTcginaDLL; pfil; pfil = pfil->pNextItem) {
        UninstallFile(pfil->fi.fd.cFileName, pfRebootRequired);
    }
    if (psdd->pfilFirstTcginaLog) {
        PFILE_INFO_LIST pfil;
        for (pfil = psdd->pfilFirstTcginaLog; pfil; pfil = pfil->pNextItem) {
            UninstallFile(pfil->fi.fd.cFileName, pfRebootRequired);
        }
    }

    /* Reset user profiles */
    if (SendMessage(psdd->hwndUninstallCheckBox2, BM_GETCHECK, 0, 0) ==
      BST_CHECKED) {
        DWORD i;
        for (i = 0; i < psdd->pui->dwUserCount; i++) {
            if (!psdd->pui->pItem[i].fUninstallable) {
                continue;
            }
            if (!UninstallUser(psdd->pui->pItem + i)) {
                return FALSE;
            }
        }
    }

    /* Remove registry entries */
    if (SendMessage(psdd->hwndUninstallCheckBox3, BM_GETCHECK, 0, 0) ==
      BST_CHECKED) {
        if (!RegRemoveKey(HKEY_LOCAL_MACHINE, TCGINA_KEY)) {
            return FALSE;
        }
        if (RegKeyIsEmpty(HKEY_LOCAL_MACHINE, TCGINA_ROOT_KEY)) {
            if (!RegRemoveKey(HKEY_LOCAL_MACHINE, TCGINA_ROOT_KEY)) {
                return FALSE;
            }
        }
    }

    return TRUE;
}

static BOOL GetComboBoxString(HWND hwnd, int iIndex, LPTSTR szDest,
  DWORD nSize)
{
    LRESULT lResult;

    if (iIndex < 0) {
        lResult = SendMessage(hwnd, CB_GETCURSEL, 0, 0);
        if (lResult == CB_ERR) return FALSE;
        iIndex = (int)lResult;
    }
    lResult = SendMessage(hwnd, CB_GETLBTEXTLEN, (WPARAM)iIndex, 0);
    if (lResult == CB_ERR) return FALSE;
    if ((DWORD)lResult >= nSize) return FALSE;
    return SendMessage(hwnd, CB_GETLBTEXT, (WPARAM)iIndex,
      (LPARAM)szDest) != CB_ERR;
}

static DWORD GetVersionFromString(LPCTSTR pszVersion)
{
    DWORD  dwMajor;
    DWORD  dwMinor;
    DWORD  dwRevision;
    LPTSTR pszTemp;

    dwMajor = dwMinor = dwRevision = 0;

    dwMajor = (DWORD)_tcstoul(pszVersion, &pszTemp, 10);
    if (pszTemp && *pszTemp == '.') {
        dwMinor = (DWORD)_tcstoul(pszTemp + 1, &pszTemp, 10);
        if (pszTemp && _istalpha(pszTemp[0])) {
            dwRevision = (tolower((char)pszTemp[0]) - ('a' - 1));
        }
    }

    return (dwMajor << 16) + (dwMinor << 8) + dwRevision;
}

static void GetWarningString(PSETUP_DIALOG_DATA psdd, DWORD dwIndex,
  LPTSTR pszWarning, DWORD nSize)
{
    BOOL  fStatusWarning;
    PFILE_INFO pfiGinaDLL;
    TCHAR szNote[256];
    TCHAR szTrueCryptWarning[256];
    TCHAR szLMHashWarning[256];

    fStatusWarning = dwIndex >= psdd->dwNewGinaDLLCount;
    pfiGinaDLL = fStatusWarning ? &psdd->fiGinaDLL : psdd->pfiNewGinaDLL + dwIndex;
    *szNote = TEXT('\0');
    *szTrueCryptWarning = TEXT('\0');
    *szLMHashWarning = TEXT('\0');

    if (fStatusWarning) {
        if (!psdd->fiTrueCrypt.fInUse) {
            _sntprintf(szTrueCryptWarning, countof(szTrueCryptWarning),
              TEXT("WARNING: TrueCrypt is not installed!\r\n")
              TEXT("(TCGINA is therefore not able to log on with an encrypted ")
              TEXT("user profile)\r\n"));
        } else if (psdd->iStatus >= TCGS_INSTALLED) {
            TCHAR szTCVersion[40];
            TCHAR szRequiredMinTCVersion[40];
            TCHAR szRequiredMaxTCVersion[40];
            DWORD dwTCVersion;
            DWORD dwRequiredMinTCVersion;
            DWORD dwRequiredMaxTCVersion;
            GetInfoItem(psdd->fiTrueCrypt.szInfo, 1, szTCVersion,
              countof(szTCVersion));
            GetInfoItem(pfiGinaDLL->szInfo, 3, szRequiredMinTCVersion,
              countof(szRequiredMinTCVersion));
            if (!GetInfoItem(pfiGinaDLL->szInfo, 4, szRequiredMaxTCVersion,
              countof(szRequiredMaxTCVersion))) {
                TC_CopyString(szRequiredMaxTCVersion,
                  countof(szRequiredMaxTCVersion), szRequiredMinTCVersion);
            }
            dwTCVersion = GetVersionFromString(szTCVersion);
            dwRequiredMinTCVersion =
              GetVersionFromString(szRequiredMinTCVersion);
            dwRequiredMaxTCVersion =
              GetVersionFromString(szRequiredMaxTCVersion);
            if (dwRequiredMinTCVersion &&
              (dwTCVersion < dwRequiredMinTCVersion ||
              dwTCVersion > dwRequiredMaxTCVersion)) {
                if (dwRequiredMinTCVersion != dwRequiredMaxTCVersion) {
                    TC_AppendString(szRequiredMinTCVersion,
                      countof(szRequiredMinTCVersion), TEXT("-"));
                    TC_AppendString(szRequiredMinTCVersion,
                      countof(szRequiredMinTCVersion), szRequiredMaxTCVersion);
                }
                _sntprintf(szTrueCryptWarning, countof(szTrueCryptWarning),
                  TEXT("WARNING: TCGINA requires TrueCrypt %s!\r\n"
                  TEXT("(TCGINA is possibly not able to log on with an encrypted ")
                  TEXT("user profile)\r\n")),
                  szRequiredMinTCVersion);
            }
        }
    } else {
        TCHAR szVersion[128];

        TC_CopyString(szVersion, countof(szVersion), TEXT("TCGINA"));
        GetComboBoxString(psdd->hwndInstallComboBox, dwIndex, szVersion,
          countof(szVersion));

        if (psdd->iStatus >= TCGS_INSTALLED && *psdd->fiGinaDLL.szInfo &&
          *pfiGinaDLL->szInfo && memcmp(psdd->fiGinaDLL.rgbHashValue,
          pfiGinaDLL->rgbHashValue, FILE_INFO_HASH_SIZE) == 0) {
            _sntprintf(szNote, countof(szNote),
              TEXT("NOTE: %s is already installed.\r\n"), szVersion);
            EnableWindow(psdd->hwndOk, FALSE);
        } else {
            EnableWindow(psdd->hwndOk, TRUE);
        }

        if (!psdd->fiTrueCrypt.fInUse) {
            _sntprintf(szTrueCryptWarning, countof(szTrueCryptWarning),
              TEXT("WARNING: TrueCrypt is not installed!\r\n")
              TEXT("(TCGINA will therefore not be able to log on with an encrypted ")
              TEXT("user profile)\r\n"));
        } else if (psdd->iStatus >= TCGS_INSTALLED) {
            TCHAR szTCVersion[40];
            TCHAR szRequiredMinTCVersion[40];
            TCHAR szRequiredMaxTCVersion[40];
            DWORD dwTCVersion;
            DWORD dwRequiredMinTCVersion;
            DWORD dwRequiredMaxTCVersion;

            GetInfoItem(psdd->fiTrueCrypt.szInfo, 1, szTCVersion,
              countof(szTCVersion));
            GetInfoItem(pfiGinaDLL->szInfo, 3, szRequiredMinTCVersion,
              countof(szRequiredMinTCVersion));
            if (!GetInfoItem(pfiGinaDLL->szInfo, 4, szRequiredMaxTCVersion,
              countof(szRequiredMaxTCVersion))) {
                TC_CopyString(szRequiredMaxTCVersion,
                  countof(szRequiredMaxTCVersion), szRequiredMinTCVersion);
            }
            dwTCVersion = GetVersionFromString(szTCVersion);
            dwRequiredMinTCVersion =
              GetVersionFromString(szRequiredMinTCVersion);
            dwRequiredMaxTCVersion =
              GetVersionFromString(szRequiredMaxTCVersion);
            if (dwRequiredMinTCVersion &&
              (dwTCVersion < dwRequiredMinTCVersion ||
              dwTCVersion > dwRequiredMaxTCVersion)) {
                if (dwRequiredMinTCVersion != dwRequiredMaxTCVersion) {
                    TC_AppendString(szRequiredMinTCVersion,
                      countof(szRequiredMinTCVersion), TEXT("-"));
                    TC_AppendString(szRequiredMinTCVersion,
                      countof(szRequiredMinTCVersion), szRequiredMaxTCVersion);
                }
                _sntprintf(szTrueCryptWarning, countof(szTrueCryptWarning),
                  TEXT("WARNING: %s requires TrueCrypt %s!\r\n")
                  TEXT("(TCGINA will possibly not be able to log on with an encrypted ")
                  TEXT("user profile)\r\n"), szVersion, szRequiredMinTCVersion);
            }
        }
    }
    if (psdd->dwNoLMHash != 1) {
        _sntprintf(szLMHashWarning, countof(szLMHashWarning),
          TEXT("WARNING: Generation of LAN Manager hashes is not disabled!\r\n")
          TEXT("(User account passwords should therefore not be used ")
          TEXT("for TrueCrypt volumes)"));
    }

    TC_CopyString(pszWarning, nSize, szNote);
    if (*szTrueCryptWarning) {
        if (*pszWarning) {
            TC_AppendString(pszWarning, nSize, TEXT("\r\n"));
        }
        TC_AppendString(pszWarning, nSize, szTrueCryptWarning);
    }
    if (*szLMHashWarning) {
        if (*pszWarning) {
            TC_AppendString(pszWarning, nSize, TEXT("\r\n"));
        }
        TC_AppendString(pszWarning, nSize, szLMHashWarning);
    }
}

static void UpdateInstallText(PSETUP_DIALOG_DATA psdd)
{
    DWORD dwIndex = (DWORD)SendMessage(psdd->hwndInstallComboBox,
      CB_GETCURSEL, 0, 0);
    if (dwIndex < psdd->dwNewGinaDLLCount) {
        TCHAR szText[1024];
        TCHAR szWarning[512];
        GetWarningString(psdd, dwIndex, szWarning, countof(szWarning));
        _sntprintf(szText, countof(szText), TEXT("%s"), szWarning);
        SetWindowText(psdd->hwndInstallText, szText);
    }
}

static void UpdateEncryptPage(PSETUP_DIALOG_DATA psdd)
{
    TCHAR   szText[512];
    BOOL    fShowOkButton;
    LRESULT lResult;

    if (psdd->pui == NULL) {
        return;
    }
    *szText = TEXT('\0');
    fShowOkButton = FALSE;

    *psdd->szSrcPath = TEXT('\0');
    *psdd->szDestPath = TEXT('\0');
    *psdd->szTrueCryptVolume = TEXT('\0');

    lResult = SendMessage(psdd->hwndEncryptComboBox2, CB_GETCURSEL, 0, 0);
    if (lResult == CB_ERR) {
        if (SendMessage(psdd->hwndEncryptComboBox2, CB_GETCOUNT, 0, 0) < 1) {
            _sntprintf(szText, countof(szText),
              TEXT("Mount a TrueCrypt volume for the encrypted user ")
              TEXT("profile!\r\n"));
        } else {
            _sntprintf(szText, countof(szText),
              TEXT("Select a destination drive for the encrypted user ")
              TEXT("profile!\r\n"));
        }
    } else {
        DWORD dwDrive;
        DWORD dwUser;
        DWORD nLen;

        for (dwDrive = 0; dwDrive < 26; dwDrive++) {
            if (*psdd->pui->rgszTCFileName[dwDrive] == TEXT('\0')) {
                continue;
            }
            if (lResult-- == 0) {
                TC_CopyString(psdd->szTrueCryptVolume, MAX_PATH,
                  psdd->pui->rgszTCFileName[dwDrive]);
                psdd->dwDrive = dwDrive;
                break;
            }
        }

        lResult = SendMessage(psdd->hwndEncryptComboBox1, CB_GETCURSEL, 0, 0);
        for (dwUser = 0; dwUser < psdd->pui->dwUserCount; dwUser++) {
            if (psdd->pui->pItem[dwUser].fEncryptedProfile) {
                continue;
            }
            if (lResult-- == 0) {
                psdd->puii = psdd->pui->pItem + dwUser;
                break;
            }
        }
        if (dwUser == psdd->pui->dwUserCount) {
            _sntprintf(szText, countof(szText),
              TEXT("ERROR: Invalid user name!"));
        } else {
            LPTSTR pszDirectory = psdd->szSrcPath;
            nLen = (DWORD)_tcslen(psdd->szTrueCryptVolume);
            if (nLen == 0) {
                _sntprintf(szText, countof(szText),
                  TEXT("ERROR: Invalid TrueCrypt drive!"));
#if 0
            /* only required for TrueCrypt version <= 4.0 */
            } else if (nLen > 3) {
                if (psdd->szTrueCryptVolume[nLen-3] == TEXT('.') &&
                  psdd->szTrueCryptVolume[nLen-2] == TEXT('.') &&
                  psdd->szTrueCryptVolume[nLen-1] == TEXT('.')) {
                    _sntprintf(szText, countof(szText),
                      TEXT("ERROR: TrueCrypt volume host filename is too long!"));
                    nLen = 0;
                }
#endif
            }
            if (nLen) {
                TC_CopyString(psdd->szSrcPath, MAX_PATH,
                    psdd->pui->pItem[dwUser].szProfileImagePath);
                if (_istalpha(pszDirectory[0]) && pszDirectory[1] == TEXT(':')) {
                  pszDirectory += 2;
                } else if (pszDirectory[0] == TEXT('\\') && pszDirectory[1] ==
                  TEXT('\\') && pszDirectory[2] != '?' &&
                  (pszDirectory[2] != '.' || pszDirectory[3] != TEXT('\\'))) {
                    pszDirectory = _tcschr(pszDirectory + 2, TEXT('\\'));
                    if (pszDirectory == TEXT('\0')) {
                        pszDirectory = psdd->szSrcPath + _tcslen(psdd->szSrcPath);
                    }
                } else {
                    _sntprintf(szText, countof(szText),
                      TEXT("ERROR: Unsupported profile image path!"));
                    nLen = 0;
                }
            }
            if (nLen) {
                TCHAR szNtUserDat[MAX_PATH];
                if (*pszDirectory == TEXT('\\')) pszDirectory++;
                psdd->szDestPath[0] = (TCHAR)(TEXT('A') + dwDrive);
                psdd->szDestPath[1] = TEXT(':');
                psdd->szDestPath[2] = TEXT('\\');
                TC_CopyString(psdd->szDestPath+3, MAX_PATH-3, pszDirectory);
                TC_CopyString(szNtUserDat, MAX_PATH, psdd->szDestPath);
                TC_AppendString(szNtUserDat, MAX_PATH, TEXT("\\NTUSER.DAT"));

                if (FileExists(szNtUserDat)) {
                    *psdd->szSrcPath = TEXT('\0');
                    _sntprintf(szText, countof(szText),
                      TEXT("TrueCrypt Volume: %s\r\n\r\n")
                      TEXT("WARNING: The TrueCrypt volume already contains a ")
                      TEXT("profile for the selected user. The user profile ")
                      TEXT("will therefore only be redirected to the ")
                      TEXT("encrypted location, but no files will be copied.")
                      TEXT(""), psdd->szTrueCryptVolume);
                    psdd->iEncryptionMethod = TCEM_REDIRECT;
                    fShowOkButton = TRUE;
                } else if (FileExists(psdd->szDestPath)) {
                    _sntprintf(szText, countof(szText),
                      TEXT("TrueCrypt Volume: %s\r\n\r\n")
                      TEXT("ERROR: The destination folder ")
                      TEXT("\"%s\" already exists!")
                      TEXT(""), psdd->szTrueCryptVolume, psdd->szDestPath);
                } else {
                    _sntprintf(szText, countof(szText),
                      TEXT("TrueCrypt Volume: %s\r\n\r\n")
                      TEXT("Press OK to copy the user profile to the ")
                      TEXT("TrueCrypt volume and to redirect the user ")
                      TEXT("profile to the encrypted location.\r\n")
                      TEXT(""), psdd->szTrueCryptVolume);
                    psdd->iEncryptionMethod = TCEM_COPY;
                    fShowOkButton = TRUE;
                }
                if (fShowOkButton) {
                    if (TC_RemoteFile(psdd->szTrueCryptVolume)) {
                        TCHAR szText2[256];
                        _sntprintf(szText2, countof(szText2),
                          TEXT("\r\nWARNING: Note that TCGINA will use the ")
                          TEXT("credentials of user \"%s\" to connect to the server.")
                          TEXT("\r\nWARNING: Note that you cannot use TCGINA LITE to logon ")
                          TEXT("with this encrypted user profile."),
                            psdd->puii->szUserName);
                        TC_AppendString(szText, countof(szText), szText2);
                    } else if (psdd->szTrueCryptVolume[1] == TEXT(':') &&
                      TC_GetDriveType(psdd->szTrueCryptVolume[0]) == DRIVE_REMOTE) {
                        TCHAR szText2[256];
                        _sntprintf(szText2, countof(szText2),
                          TEXT("\r\nWARNING: Note that you must map the network share ")
                          TEXT("to drive letter %c: before logon (see TCGINA.PDF).")
                          TEXT("\r\nWARNING: Note that you cannot use TCGINA LITE to logon ")
                          TEXT("with this encrypted user profile."),
                            _totupper(psdd->szTrueCryptVolume[0]));
                        TC_AppendString(szText, countof(szText), szText2);
                    } else if (psdd->szTrueCryptVolume[1] == TEXT(':')) {
                        int nDriveNo = _totupper(psdd->szTrueCryptVolume[0]) - TEXT('A');
                        if (nDriveNo < 26 && *psdd->pui->rgszTCFileName[nDriveNo]) {
                            TCHAR szText2[256];
                            _sntprintf(szText2, countof(szText2),
                            TEXT("\r\nWARNING: Note that you must mount the TrueCrypt volume ")
                            TEXT("\"%s\" with drive letter %c: before logon (see TCGINA.PDF).")
                            TEXT("\r\nWARNING: Note that you cannot use TCGINA LITE to logon ")
                            TEXT("with this encrypted user profile."),
                                psdd->pui->rgszTCFileName[nDriveNo],
                                _totupper(psdd->szTrueCryptVolume[0]));
                            TC_AppendString(szText, countof(szText), szText2);
                        }
                    }
                }
            }
        }
    }

    SetWindowText(psdd->hwndEncryptText, szText);
    EnableWindow(psdd->hwndOk, fShowOkButton);
}

static BOOL EncryptUserProfile(PSETUP_DIALOG_DATA psdd)
{
    SHFILEOPSTRUCT shFileOp;
    TCHAR szUndoPaths[MAX_PATH+1];
    TCHAR szProfileKey[MAX_PATH];
    TCHAR szTcginaKey[2*MAX_PATH];
    DWORD cbSize;
    DWORD dwType;

    SetCursor(psdd->hWaitCursor);

    /* Copy profile */
    if (*psdd->szSrcPath) {
        TCHAR                szSrcPaths[MAX_PATH+1];
        TCHAR                szDestPaths[MAX_PATH+1];
        LPTSTR               pszTemp;

        /* szUndoPaths is doubly TEXT('\0') terminated */
        memset(szUndoPaths, '\0', sizeof(szUndoPaths));
        /* Create parent folders */
        pszTemp = _tcschr(psdd->szDestPath+3, TEXT('\\'));
        while (pszTemp) {
            *pszTemp = TEXT('\0');
            if (CreateDirectory(psdd->szDestPath, NULL)) {
                if (*szUndoPaths == TEXT('\0')) {
                    TC_CopyString(szUndoPaths, MAX_PATH, psdd->szDestPath);
                }
            }
            *pszTemp = TEXT('\\');
            pszTemp = _tcschr(pszTemp + 1, TEXT('\\'));
        }
        if (*szUndoPaths == TEXT('\0')) {
            TC_CopyString(szUndoPaths, MAX_PATH, psdd->szDestPath);
        }

        /* Copy files */
        memset(szSrcPaths, '\0', sizeof(szSrcPaths));
        TC_CopyString(szSrcPaths, countof(szSrcPaths), psdd->szSrcPath);
        memset(szDestPaths, '\0', sizeof(szDestPaths));
        TC_CopyString(szDestPaths, countof(szDestPaths), psdd->szDestPath);
        memset(&shFileOp, '\0', sizeof(shFileOp));
        shFileOp.wFunc = FO_COPY;
        shFileOp.pFrom = szSrcPaths;
        shFileOp.pTo = szDestPaths;
        shFileOp.fFlags = FOF_NOCONFIRMATION;
        shFileOp.fAnyOperationsAborted = FALSE;
        if (SHFileOperation(&shFileOp) != 0 ||
          shFileOp.fAnyOperationsAborted == TRUE) {
            goto error;
        }
        CopyPathSecurity(psdd->szSrcPath, psdd->szDestPath);
    }

    /* Get registry value type */
    TC_CopyString(szProfileKey, MAX_PATH, PROFILELIST_KEY);
    TC_AppendString(szProfileKey, MAX_PATH, TEXT("\\"));
    TC_AppendString(szProfileKey, MAX_PATH, psdd->puii->szSID);
    cbSize = 0;
    dwType = REG_NONE;
    RegGetValueData(HKEY_LOCAL_MACHINE, szProfileKey,
      TEXT("ProfileImagePath"), &dwType, NULL, &cbSize);
    if (dwType != REG_SZ && dwType != REG_EXPAND_SZ) {
        goto error;
    }

    /* Create uninstall info */
    TC_CopyString(szTcginaKey, countof(szTcginaKey), TCGINA_KEY);
    TC_AppendString(szTcginaKey, countof(szTcginaKey), TEXT("\\"));
    TC_AppendString(szTcginaKey, countof(szTcginaKey), psdd->puii->szUserName);

    if (!RegSetString(HKEY_LOCAL_MACHINE, szTcginaKey, TEXT("SID"),
      REG_SZ, psdd->puii->szSID)) {
        goto error;
    }
    if (!RegCopyValue(HKEY_LOCAL_MACHINE, szProfileKey,
      TEXT("ProfileImagePath"), HKEY_LOCAL_MACHINE, szTcginaKey,
      TEXT("ProfileImagePath"))) {
        goto error;
    }
    if (!RegSetString(HKEY_LOCAL_MACHINE, szTcginaKey,
      TEXT("EncryptedProfileImagePath"), dwType, psdd->szDestPath)) {
        goto error;
    }

    /*
     * Remove 'TCDrive' registry value because the drive letter is
     * taken from the registry value 'ProfileImagePath'
     */
    RegRemoveValue(HKEY_LOCAL_MACHINE, szTcginaKey,
      TEXT("TCDrive"));
    /* Create TCGINA registry values */
    if (!RegSetString(HKEY_LOCAL_MACHINE, szTcginaKey, TEXT("TCFileName"),
      REG_SZ, psdd->szTrueCryptVolume)) {
        goto error;
    }

    /* Redirect profile */
    if (!RegSetString(HKEY_LOCAL_MACHINE, szProfileKey,
      TEXT("ProfileImagePath"), dwType, psdd->szDestPath)) {
        goto error;
    }

    SetCursor(psdd->hArrowCursor);
    return TRUE;

error:
    UninstallUser(psdd->puii);
    if (*psdd->szSrcPath) {
        memset(&shFileOp, '\0', sizeof(shFileOp));
        shFileOp.wFunc = FO_DELETE;
        shFileOp.pFrom = szUndoPaths;
        shFileOp.fFlags = FOF_NOCONFIRMATION;
        shFileOp.fAnyOperationsAborted = FALSE;
        SHFileOperation(&shFileOp);
        UpdateEncryptPage(psdd);
    }
    SetCursor(psdd->hArrowCursor);
    return FALSE;

}

static BOOL ResetUserProfile(PSETUP_DIALOG_DATA psdd)
{
    LRESULT lResult;
    DWORD   dwUser;

    lResult = SendMessage(psdd->hwndResetComboBox, CB_GETCURSEL, 0, 0);
    for (dwUser = 0; dwUser < psdd->pui->dwUserCount; dwUser++) {
        if (!psdd->pui->pItem[dwUser].fUninstallable) {
            continue;
        }
        if (lResult-- == 0) {
            return UninstallUser(psdd->pui->pItem + dwUser);
        }
    }

    return FALSE;
}

static void UpdateTrueCryptComboBox(PSETUP_DIALOG_DATA psdd)
{
    TCHAR   szDrive[40];
    TCHAR   szTemp[40];
    LRESULT lResult;
    DWORD i;

    if (psdd->pui == NULL) {
        return;
    }

    if (!GetComboBoxString(psdd->hwndEncryptComboBox2, -1, szDrive,
      countof(szDrive))) {
        *szDrive = TEXT('\0');
    }

    SendMessage(psdd->hwndEncryptComboBox2, CB_RESETCONTENT, 0, 0);
    if (UpdateTCFileNames(psdd->pui)) {
        szTemp[1] = TEXT(':');
        szTemp[2] = TEXT('\0');
        for (i = 0; i < 26; i++) {
            if (*psdd->pui->rgszTCFileName[i]) {
                szTemp[0] = (TCHAR)(TEXT('A') + i);
                SendMessage(psdd->hwndEncryptComboBox2, CB_ADDSTRING, 0,
                (LPARAM) szTemp);
            }
        }
    }

    lResult = SendMessage(psdd->hwndEncryptComboBox2, CB_FINDSTRINGEXACT,
      (WPARAM)-1, (LPARAM)szDrive);
    SendMessage(psdd->hwndEncryptComboBox2, CB_SETCURSEL, lResult, 0);

    if (psdd->iPage == PG_ENCRYPT_USER_PROFILE) {
        UpdateEncryptPage(psdd);
    }
}

static void UpdateResetPage(PSETUP_DIALOG_DATA psdd)
{
    UNREFERENCED_PARAMETER(psdd);
}

static void SelectPage(PSETUP_DIALOG_DATA psdd, int iPage)
{
    int fShowEdit = FALSE;
    int fShowOkButton = FALSE;

    if (iPage == PG_NOCHANGE) {
        iPage = (int)SendMessage(psdd->hwndList, LB_GETCURSEL, 0, 0);
    }
    if (psdd->iPage == iPage) return;
    SetFocus(psdd->hwndList);

    ShowWindow(psdd->hwndInstallLabel, SW_HIDE);
    ShowWindow(psdd->hwndInstallComboBox, SW_HIDE);
    ShowWindow(psdd->hwndInstallText, SW_HIDE);
    ShowWindow(psdd->hwndEncryptLabel1, SW_HIDE);
    ShowWindow(psdd->hwndEncryptComboBox1, SW_HIDE);
    ShowWindow(psdd->hwndEncryptLabel2, SW_HIDE);
    ShowWindow(psdd->hwndEncryptComboBox2, SW_HIDE);
    ShowWindow(psdd->hwndEncryptText, SW_HIDE);
    ShowWindow(psdd->hwndResetLabel, SW_HIDE);
    ShowWindow(psdd->hwndResetComboBox, SW_HIDE);
    ShowWindow(psdd->hwndResetText, SW_HIDE);
    ShowWindow(psdd->hwndUninstallCheckBox1, SW_HIDE);
    ShowWindow(psdd->hwndUninstallCheckBox2, SW_HIDE);
    ShowWindow(psdd->hwndUninstallCheckBox3, SW_HIDE);
    EnableWindow(psdd->hwndOk, FALSE);

    switch(iPage) {
    case PG_INSTALL:
        if (PendingFileRenameOperationsContains(TEXT("TCGINA"))) {
            SetWindowText(psdd->hwndText,
              TEXT("ERROR: Cannot install TCGINA.\r\n\r\n")
              TEXT("Another program requires the computer to reboot!"));
            fShowEdit = TRUE;
        } else if (psdd->dwNewGinaDLLCount == 0) {
            SetWindowText(psdd->hwndText,
              TEXT("ERROR: Cannot find installation files.\r\n"));
            fShowEdit = TRUE;
        } else {
            UpdateInstallText(psdd);
            ShowWindow(psdd->hwndInstallLabel, SW_SHOWNORMAL);
            ShowWindow(psdd->hwndInstallComboBox, SW_SHOWNORMAL);
            ShowWindow(psdd->hwndInstallText, SW_SHOWNORMAL);
        }
        break;

    case PG_ENCRYPT_USER_PROFILE:
        if (psdd->iStatus < TCGS_INSTALLED) {
            SetWindowText(psdd->hwndText,
              TEXT("ERROR: TCGINA is not installed!"));
            fShowEdit = TRUE;
        } else if (psdd->pui == NULL) {
            SetWindowText(psdd->hwndText,
              TEXT("ERROR: Cannot enumerate user accounts!"));
            fShowEdit = TRUE;
        } else if (psdd->pui->dwUnencryptedUsers == 0) {
            SetWindowText(psdd->hwndText,
              TEXT("No encryptable user profiles found.\r\n\r\n")
              TEXT("(1) Create a new user account\r\n\r\n")
              TEXT("(2) Log off and log on to newly created user account\r\n\r\n")
              TEXT("(3) Log off and log on again to this user account\r\n\r\n"));
            fShowEdit = TRUE;
        } else {
            TC_HANDLE hTC = TC_Open(0);
            if (hTC == NULL) {
                SetWindowText(psdd->hwndText,
                  TEXT("ERROR: Cannot access TrueCrypt (Required Version: ")
                  TEXT(PROJECT_REQUIRED_TC_VERSION)
                  TEXT(")!"));
                fShowEdit = TRUE;
            } else {
                TC_Close(hTC);
                UpdateEncryptPage(psdd);
                ShowWindow(psdd->hwndEncryptLabel1, SW_SHOWNORMAL);
                ShowWindow(psdd->hwndEncryptComboBox1, SW_SHOWNORMAL);
                ShowWindow(psdd->hwndEncryptLabel2, SW_SHOWNORMAL);
                ShowWindow(psdd->hwndEncryptComboBox2, SW_SHOWNORMAL);
                ShowWindow(psdd->hwndEncryptText, SW_SHOWNORMAL);
            }
        }
        break;

    case PG_RESET_USER_PROFILE:
        if (psdd->pui == NULL) {
            SetWindowText(psdd->hwndText,
              TEXT("ERROR: Cannot enumerate user accounts!"));
            fShowEdit = TRUE;
        } else if (psdd->pui->dwUninstallableUsers == 0) {
            SetWindowText(psdd->hwndText,
              TEXT("No encrypted user profile found which can be ")
              TEXT("reset to its original unencrypted location."));
            fShowEdit = TRUE;
        } else {
            UpdateResetPage(psdd);
            ShowWindow(psdd->hwndResetLabel, SW_SHOWNORMAL);
            ShowWindow(psdd->hwndResetComboBox, SW_SHOWNORMAL);
            ShowWindow(psdd->hwndResetText, SW_SHOWNORMAL);
            SetWindowText(psdd->hwndResetText,
              TEXT("Press OK to use the original unencrypted location for ")
              TEXT("the selected user. Note that no files will ")
              TEXT("be copied."));
            fShowOkButton = TRUE;
        }
        break;

    case PG_UNINSTALL:
        if (psdd->iStatus == TCGS_NOT_AVAILABLE) {
            SetWindowText(psdd->hwndText,
              TEXT("There are neither TCGINA files nor TCGINA registry ")
              TEXT("entries on this computer.\r\n"));
            fShowEdit = TRUE;
        } else {
            ShowWindow(psdd->hwndUninstallCheckBox1, SW_SHOWNORMAL);
            ShowWindow(psdd->hwndUninstallCheckBox2, SW_SHOWNORMAL);
            ShowWindow(psdd->hwndUninstallCheckBox3, SW_SHOWNORMAL);
            if (psdd->pui && psdd->pui->dwUninstallableUsers) {
                EnableWindow(psdd->hwndUninstallCheckBox2, TRUE);
            } else {
                SendMessage(psdd->hwndUninstallCheckBox2, BM_SETCHECK,
                  BST_UNCHECKED, 0);
                EnableWindow(psdd->hwndUninstallCheckBox2, FALSE);
            }
            if (psdd->fRegistryKeyExists) {
                EnableWindow(psdd->hwndUninstallCheckBox3, TRUE);
            } else {
                SendMessage(psdd->hwndUninstallCheckBox3, BM_SETCHECK,
                  BST_UNCHECKED, 0);
                EnableWindow(psdd->hwndUninstallCheckBox3, FALSE);
            }
            fShowOkButton = TRUE;
        }
        break;
    case PG_STATUS:
    default: {
        TCHAR szText[1024];
        TCHAR szTcginaStatus[128];
        TCHAR szWarning[512];

        switch(psdd->iStatus) {
        case TCGS_NOT_AVAILABLE:
            _sntprintf(szTcginaStatus, countof(szTcginaStatus),
              TEXT("STATUS: TCGINA is not installed."));
            break;
        case TCGS_PASSIVE:
            _sntprintf(szTcginaStatus, countof(szTcginaStatus),
              TEXT("STATUS: TCGINA is not completely uninstalled."));
            break;
        case TCGS_INSTALLED:
            _sntprintf(szTcginaStatus, countof(szTcginaStatus),
              TEXT("STATUS: %s is installed (but requires the computer to ")
              TEXT("reboot).\r\n"), psdd->szTcginaVersion);
            break;
        case TCGS_RUNNING:
            _sntprintf(szTcginaStatus, countof(szTcginaStatus),
              TEXT("STATUS: %s is installed.\r\n"),
              psdd->szTcginaVersion);
            break;
        }

        *szWarning = '\0';
        if (psdd->iStatus >= TCGS_INSTALLED) {
            GetWarningString(psdd, (DWORD)-1, szWarning,
              countof(szWarning));
        }

        iPage = 0;
        _sntprintf(szText, countof(szText), TEXT("%s\r\n%s"), szTcginaStatus,
          szWarning);
        SetWindowText(psdd->hwndText, szText);
        fShowEdit = TRUE;
        break; }
    }

    psdd->iPage = iPage;
    SendMessage(psdd->hwndList, LB_SETCURSEL, iPage, 0);

    ShowWindow(psdd->hwndText, fShowEdit ? SW_SHOWNORMAL : SW_HIDE);
    if (fShowOkButton) {
        EnableWindow(psdd->hwndOk, TRUE);
    }
}

INT_PTR CALLBACK SetupDialogProc(HWND hwndDlg, UINT uMsg,
  WPARAM wParam, LPARAM lParam)
{
    PSETUP_DIALOG_DATA psdd;

    switch (uMsg) {
    case WM_INITDIALOG: {
        TCHAR  szVersions[1024];
        TCHAR  szFileName[MAX_PATH];
        LPTSTR pszFilePart;
        LPTSTR pszTemp;
        DWORD  dwTemp;

        psdd = (PSETUP_DIALOG_DATA) lParam;

        psdd->hwndDlg    = hwndDlg;
        psdd->hwndOk     = GetDlgItem(hwndDlg, IDOK);
        psdd->hwndCancel = GetDlgItem(hwndDlg, IDCANCEL);
        psdd->hwndList   = GetDlgItem(hwndDlg, IDC_LIST);
        psdd->hwndText   = GetDlgItem(hwndDlg, IDC_TEXT);

        psdd->hwndInstallLabel = GetDlgItem(hwndDlg, IDC_INSTALL_LABEL);
        psdd->hwndInstallComboBox = GetDlgItem(hwndDlg, IDC_INSTALL_COMBOBOX);
        psdd->hwndInstallText = GetDlgItem(hwndDlg, IDC_INSTALL_TEXT);

        psdd->hwndEncryptLabel1 = GetDlgItem(hwndDlg, IDC_ENCRYPT_LABEL1);
        psdd->hwndEncryptComboBox1 = GetDlgItem(hwndDlg, IDC_ENCRYPT_COMBOBOX1);
        psdd->hwndEncryptLabel2 = GetDlgItem(hwndDlg, IDC_ENCRYPT_LABEL2);
        psdd->hwndEncryptComboBox2 = GetDlgItem(hwndDlg, IDC_ENCRYPT_COMBOBOX2);
        psdd->hwndEncryptText = GetDlgItem(hwndDlg, IDC_ENCRYPT_TEXT);
        if (psdd->pui && psdd->pui->dwUnencryptedUsers) {
            DWORD i;
            for (i = 0; i < psdd->pui->dwUserCount; i++) {
                if (!psdd->pui->pItem[i].fEncryptedProfile) {
                    SendMessage(psdd->hwndEncryptComboBox1, CB_ADDSTRING, 0,
                      (LPARAM) psdd->pui->pItem[i].szUserName);
                }
            }
            SendMessage(psdd->hwndEncryptComboBox1, CB_SETCURSEL, 0, 0);
            UpdateTrueCryptComboBox(psdd);
        }

        psdd->hwndResetLabel = GetDlgItem(hwndDlg, IDC_RESET_LABEL);
        psdd->hwndResetComboBox = GetDlgItem(hwndDlg, IDC_RESET_COMBOBOX);
        psdd->hwndResetText = GetDlgItem(hwndDlg, IDC_RESET_TEXT);
        if (psdd->pui && psdd->pui->dwUninstallableUsers) {
            DWORD i;
            for (i = 0; i < psdd->pui->dwUserCount; i++) {
                if (psdd->pui->pItem[i].fUninstallable) {
                    SendMessage(psdd->hwndResetComboBox, CB_ADDSTRING, 0,
                      (LPARAM) psdd->pui->pItem[i].szUserName);
                }
            }
            SendMessage(psdd->hwndResetComboBox, CB_SETCURSEL, 0, 0);
        }

        psdd->hwndUninstallCheckBox1 = GetDlgItem(hwndDlg, IDC_UNINSTALL_CHECKBOX1);
        psdd->hwndUninstallCheckBox2 = GetDlgItem(hwndDlg, IDC_UNINSTALL_CHECKBOX2);
        psdd->hwndUninstallCheckBox3 = GetDlgItem(hwndDlg, IDC_UNINSTALL_CHECKBOX3);
        SendMessage(psdd->hwndUninstallCheckBox1, BM_SETCHECK,
          (psdd->pfilFirstTcginaLog != NULL || psdd->fGinaDLLIsTcginaDLL ||
          psdd->pfilFirstTcginaDLL != NULL) ? BST_CHECKED : BST_UNCHECKED, 0);
        EnableWindow(psdd->hwndUninstallCheckBox1, FALSE);

        SendMessage(psdd->hwndList, LB_ADDSTRING, 0,
          (LPARAM)TEXT("Status"));
        SendMessage(psdd->hwndList, LB_ADDSTRING, 0,
          (LPARAM)TEXT("Install TCGINA"));
        SendMessage(psdd->hwndList, LB_ADDSTRING, 0,
          (LPARAM)TEXT("Encrypt User Profile"));
        SendMessage(psdd->hwndList, LB_ADDSTRING, 0,
          (LPARAM)TEXT("Reset User Profile"));
        SendMessage(psdd->hwndList, LB_ADDSTRING, 0,
          (LPARAM)TEXT("Uninstall TCGINA"));
        GetPrivateProfileString(TEXT("TCGINA Files"), NULL, TEXT(""),
          szVersions, countof(szVersions), GetIniFileName());
        dwTemp = 0;
        for (pszTemp = szVersions; *pszTemp; pszTemp += _tcslen(pszTemp) + 1) {
            dwTemp++;
        }
        psdd->pfiNewGinaDLL = calloc(dwTemp+1, sizeof(FILE_INFO));
        if (psdd->pfiNewGinaDLL == 0) {
            *szVersions = TEXT('\0');
        }
        GetModuleFileName(NULL, szFileName, MAX_PATH);
        pszFilePart = _tcsrchr(szFileName, TEXT('\\'));
        pszFilePart = pszFilePart ? pszFilePart+1 : szFileName;
        for (pszTemp = szVersions; *pszTemp; pszTemp += _tcslen(pszTemp) + 1) {
            GetPrivateProfileString(TEXT("TCGINA Files"), pszTemp, TEXT(""),
              pszFilePart,
              MAX_PATH-(DWORD)(pszFilePart-szFileName)/sizeof(TCHAR),
              GetIniFileName());
            GetFileInfo(psdd->pfiNewGinaDLL + psdd->dwNewGinaDLLCount,
              szFileName);
            if (IsTcginaDLL(psdd->pfiNewGinaDLL + psdd->dwNewGinaDLLCount)
#ifdef DISPLAY_ONLY_KNOWN_TCGINA_FILES
              && *psdd->pfiNewGinaDLL[psdd->dwNewGinaDLLCount].szInfo
#endif
              ) {
                psdd->dwNewGinaDLLCount++;
                SendMessage(psdd->hwndInstallComboBox, CB_ADDSTRING, 0,
                (LPARAM) pszTemp);
            }
        }
        if (psdd->dwNewGinaDLLCount) {
            SendMessage(psdd->hwndInstallComboBox, CB_SETCURSEL, 0, 0);
        }

        if (*psdd->szCaption) {
            SetWindowText(hwndDlg, psdd->szCaption);
        } else {
            GetWindowText(hwndDlg, psdd->szCaption, countof(psdd->szCaption));
        }
        psdd->hWaitCursor = LoadCursor(NULL, IDC_WAIT);
        psdd->hArrowCursor = LoadCursor(NULL, IDC_ARROW);
        SetWindowUserData(hwndDlg, psdd);

        SelectPage(psdd, PG_NOCHANGE);

        return TRUE; }

    case WM_COMMAND:
        psdd = (PSETUP_DIALOG_DATA)GetWindowUserData(hwndDlg);
        switch (HIWORD(wParam)) {
        case LBN_SELCHANGE:
            if (LOWORD(wParam) == IDC_LIST) {
                SelectPage(psdd, PG_NOCHANGE);
                return TRUE;
            } else if (LOWORD(wParam) == IDC_INSTALL_COMBOBOX) {
                UpdateInstallText(psdd);
                return TRUE;
            } else if (LOWORD(wParam) == IDC_ENCRYPT_COMBOBOX1) {
                UpdateEncryptPage(psdd);
                return TRUE;
            } else if (LOWORD(wParam) == IDC_ENCRYPT_COMBOBOX2) {
                UpdateEncryptPage(psdd);
                return TRUE;
            }
        }
        switch (wParam) {
        case IDOK: {
            switch (psdd->iPage) {
            case PG_INSTALL: {
                TCHAR szMsg[512];
                TCHAR szVersion[128];
                TC_CopyString(szVersion, countof(szVersion), TEXT("TCGINA"));
                GetComboBoxString(psdd->hwndInstallComboBox, -1, szVersion,
                  countof(szVersion));
                if (!Install(psdd)) {
                    _sntprintf(szMsg, countof(szMsg),
                      TEXT("ERROR: Cannot install %s!"), szVersion);
                    MessageBox(hwndDlg, szMsg, psdd->szCaption, MB_OK|MB_ICONSTOP);
                    return TRUE;
                } else {
                    _sntprintf(szMsg, countof(szMsg),
                      TEXT("%s has been successfully installed.\r\n")
                      TEXT("The TCGINA DLL will be started with the next computer reboot.\r\n\r\n")
                      TEXT("Note: It is recommendable that you run TrueCrypt with each encrypted user profile\r\n")
                      TEXT("and that you disable all auto-dismount events (Settings -> Preferences)."),
                      szVersion, szVersion);
                    MessageBox(hwndDlg, szMsg, psdd->szCaption, MB_OK|MB_ICONINFORMATION);
                }
                break; }
            case PG_ENCRYPT_USER_PROFILE: {
                TCHAR szMsg[1024];
                if (!EncryptUserProfile(psdd)) {
                    _sntprintf(szMsg, countof(szMsg),
                      TEXT("ERROR: Cannot encrypt user profile!"));
                    MessageBox(hwndDlg, szMsg, psdd->szCaption, MB_OK|MB_ICONSTOP);
                    return TRUE;
                } else {
                    switch(psdd->iEncryptionMethod) {
                    //case TCEM_COPY:
                    default:
                        _sntprintf(szMsg, countof(szMsg),
                          TEXT("The user profile has been successfully encrypted.\n\n")
                          TEXT("Note that the original files of the unencrypted user profile are still available, ")
                          TEXT("and it is up to you to erase these files securely (e. g. with Eraser).\n\n")
                          TEXT("Storage location of the unencrypted user profile:\n\n%s"),
                            psdd->puii->szProfileImagePath);
                        break;
                    case TCEM_REDIRECT:
                        _sntprintf(szMsg, countof(szMsg),
                          TEXT("The user profile has been successfully redirected to the encrypted location.\n\n")
                          TEXT("Note that the files of the unencrypted user profile are still available, ")
                          TEXT("and it is up to you to erase these files securely (e. g. with Eraser).\n\n")
                          TEXT("Storage location of the unencrypted user profile:\n\n%s"),
                            psdd->puii->szProfileImagePath);
                        break;
                    }
                    MessageBox(hwndDlg, szMsg, psdd->szCaption, MB_OK|MB_ICONINFORMATION);
                }
                break; }
            case PG_RESET_USER_PROFILE: {
                TCHAR szMsg[512];
                if (!ResetUserProfile(psdd)) {
                    _sntprintf(szMsg, countof(szMsg),
                      TEXT("ERROR: Cannot reset user profile!"));
                    MessageBox(hwndDlg, szMsg, psdd->szCaption, MB_OK|MB_ICONSTOP);
                    return TRUE;
                } else {
                    _sntprintf(szMsg, countof(szMsg),
                      TEXT("The user profile has been successfully reset."));
                    MessageBox(hwndDlg, szMsg, psdd->szCaption, MB_OK|MB_ICONINFORMATION);
                }
                break; }
            case PG_UNINSTALL: {
                TCHAR szMsg[512];
                BOOL  fRebootRequired = FALSE;
                if (!Uninstall(psdd, &fRebootRequired)) {
                    _sntprintf(szMsg, countof(szMsg),
                      TEXT("ERROR: Cannot uninstall TCGINA!"));
                    MessageBox(hwndDlg, szMsg, psdd->szCaption, MB_OK|MB_ICONSTOP);
                    return TRUE;
                } else {
                    if (fRebootRequired) {
                        _sntprintf(szMsg, countof(szMsg),
                          TEXT("TCGINA has been successfully removed.\r\n")
                          TEXT("Remaining TCGINA files will be removed with the next computer reboot."));
                    } else {
                        _sntprintf(szMsg, countof(szMsg),
                          TEXT("TCGINA has been successfully removed."));
                    }
                    MessageBox(hwndDlg, szMsg, psdd->szCaption, MB_OK|MB_ICONINFORMATION);
                }
                break; }
            default:
                return TRUE;
            }
            /* fall through */ }

        case IDCANCEL: {
            free(psdd->pfiNewGinaDLL);
            EndDialog(hwndDlg, LOWORD(wParam));
            return TRUE; }
        }
        break;

    case WM_DEVICECHANGE:
        switch (wParam) {
        case DBT_DEVICEARRIVAL:
        case DBT_DEVICEREMOVECOMPLETE:
            psdd = (PSETUP_DIALOG_DATA)GetWindowUserData(hwndDlg);
            UpdateTrueCryptComboBox(psdd);
            break;
        }
        return FALSE;
    }

    return FALSE;
}
